@@ -1,10 +1,18 @@
+from typing import List, TYPE_CHECKING, overload, Tuple, Optional, Iterable, TypeVar, Any
 import logging
 import threading
 
 from .constrained_frontend import ConstrainedFrontend
 
+if TYPE_CHECKING:
+    from ..ast.bv import BV
+    from ..ast.bool import Bool
+    from ..ast.fp import FP
+
 l = logging.getLogger("claripy.frontends.full_frontend")
 
+T = TypeVar("T", bound="FullFrontend")
+
 
 class FullFrontend(ConstrainedFrontend):
     _model_hook = None
@@ -87,12 +95,12 @@ class FullFrontend(ConstrainedFrontend):
     # Constraint management
     #
 
-    def add(self, constraints):
+    def add(self, constraints: List["Bool"]) -> List["Bool"]:
         to_add = ConstrainedFrontend.add(self, constraints)
         self._to_add += to_add
         return to_add
 
-    def simplify(self):
+    def simplify(self) -> List["Bool"]:
         ConstrainedFrontend.simplify(self)
 
         # TODO: should we do this?
@@ -101,7 +109,7 @@ class FullFrontend(ConstrainedFrontend):
 
         return self.constraints
 
-    def check_satisfiability(self, extra_constraints=(), exact=None):
+    def check_satisfiability(self, extra_constraints=(), exact=None) -> bool:
         try:
             return self._solver_backend.check_satisfiability(
                 extra_constraints=extra_constraints, solver=self._get_solver(), model_callback=self._model_hook
@@ -109,7 +117,7 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as e:
             raise ClaripyFrontendError("Backend error during solve") from e
 
-    def satisfiable(self, extra_constraints=(), exact=None):
+    def satisfiable(self, extra_constraints: Iterable["Bool"] = (), exact: Optional[bool] = None) -> bool:
         try:
             return self._solver_backend.satisfiable(
                 extra_constraints=extra_constraints, solver=self._get_solver(), model_callback=self._model_hook
@@ -117,7 +125,22 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as e:
             raise ClaripyFrontendError("Backend error during solve") from e
 
-    def eval(self, e, n, extra_constraints=(), exact=None):
+    @overload
+    def eval(
+        self, e: "BV", n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> Tuple[int, ...]: ...
+
+    @overload
+    def eval(
+        self, e: "Bool", n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> Tuple[bool, ...]: ...
+
+    @overload
+    def eval(
+        self, e: "FP", n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> Tuple[float, ...]: ...
+
+    def eval(self, e, n, extra_constraints=(), exact=None) -> Tuple[Any, ...]:
         if not self.satisfiable(extra_constraints=extra_constraints):
             raise UnsatError("unsat")
 
@@ -134,6 +157,23 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as exc:
             raise ClaripyFrontendError("Backend error during eval") from exc
 
+    # this is technically wrong but we cannot do better because python doesn't have associated types
+    # if you need another overload, add one
+    @overload
+    def batch_eval(
+        self, exprs: Iterable["BV"], n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional[bool] = ...
+    ) -> List[Tuple[int, ...]]: ...
+
+    @overload
+    def batch_eval(
+        self, exprs: Iterable["Bool"], n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional[bool] = ...
+    ) -> List[Tuple[bool, ...]]: ...
+
+    @overload
+    def batch_eval(
+        self, exprs: Iterable["FP"], n: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional[bool] = ...
+    ) -> List[Tuple[float, ...]]: ...
+
     def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
         if not self.satisfiable(extra_constraints=extra_constraints):
             raise UnsatError("unsat")
@@ -149,12 +189,28 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as e:
             raise ClaripyFrontendError("Backend error during batch_eval") from e
 
+    @overload
+    def max(
+        self, e: "BV", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> int: ...
+
+    @overload
+    def max(
+        self, e: "Bool", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> bool: ...
+
+    @overload
+    def max(
+        self, e: "FP", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> float: ...
+
     def max(self, e, extra_constraints=(), signed=False, exact=None):
         if not self.satisfiable(extra_constraints=extra_constraints):
             raise UnsatError("Unsat during _max()")
 
         l.debug("Frontend.max() with %d extra_constraints", len(extra_constraints))
 
+        # pylint: disable=unsubscriptable-object
         two = self.eval(e, 2, extra_constraints=extra_constraints)
         if len(two) == 0:
             raise UnsatError("unsat during max()")
@@ -176,12 +232,28 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as exc:
             raise ClaripyFrontendError("Backend error during max") from exc
 
+    @overload
+    def min(
+        self, e: "BV", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> int: ...
+
+    @overload
+    def min(
+        self, e: "Bool", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> bool: ...
+
+    @overload
+    def min(
+        self, e: "FP", extra_constraints: Tuple["Bool", ...] = ..., signed: bool = ..., exact: Optional["Bool"] = ...
+    ) -> float: ...
+
     def min(self, e, extra_constraints=(), signed=False, exact=None):
         if not self.satisfiable(extra_constraints=extra_constraints):
             raise UnsatError("Unsat during _min()")
 
         l.debug("Frontend.min() with %d extra_constraints", len(extra_constraints))
 
+        # pylint: disable=unsubscriptable-object
         two = self.eval(e, 2, extra_constraints=extra_constraints)
         if len(two) == 0:
             raise UnsatError("unsat during min()")
@@ -203,6 +275,21 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as exc:
             raise ClaripyFrontendError("Backend error during min") from exc
 
+    @overload
+    def solution(
+        self, e: "BV", v: int, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> bool: ...
+
+    @overload
+    def solution(
+        self, e: "Bool", v: bool, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> bool: ...
+
+    @overload
+    def solution(
+        self, e: "FP", v: float, extra_constraints: Tuple["Bool", ...] = ..., exact: Optional["Bool"] = ...
+    ) -> bool: ...
+
     def solution(self, e, v, extra_constraints=(), exact=None):
         try:
             return self._solver_backend.solution(
@@ -211,29 +298,13 @@ class FullFrontend(ConstrainedFrontend):
         except BackendError as exc:
             raise ClaripyFrontendError("Backend error during solution") from exc
 
-    def is_true(self, e, extra_constraints=(), exact=None):
+    def is_true(self, e: "Bool", extra_constraints: Tuple["Bool", ...] = (), exact: Optional[bool] = None) -> bool:
         return e.is_true()
-        # try:
-        #   return self._solver_backend.is_true(
-        #       e, extra_constraints=extra_constraints,
-        #       solver=self._get_solver(), model_callback=self._model_hook
-        #   )
-        # except BackendError:
-        #   e_type, value, traceback = sys.exc_info()
-        #   raise ClaripyFrontendError, "Backend error during _is_true: %s('%s')" % (str(e_type), str(value)), traceback
 
-    def is_false(self, e, extra_constraints=(), exact=None):
+    def is_false(self, e: "Bool", extra_constraints: Tuple["Bool", ...] = (), exact: Optional[bool] = None) -> bool:
         return e.is_false()
-        # try:
-        #   return self._solver_backend.is_false(
-        #       e, extra_constraints=extra_constraints,
-        #       solver=self._get_solver(), model_callback=self._model_hook
-        #   )
-        # except BackendError:
-        #   e_type, value, traceback = sys.exc_info()
-        #   raise ClaripyFrontendError, "Backend error during _is_false: %s('%s')" % (str(e_type), str(value)), traceback
 
-    def unsat_core(self, extra_constraints=()):
+    def unsat_core(self, extra_constraints: Tuple["Bool", ...] = ()) -> Iterable["Bool"]:
         if self.satisfiable(extra_constraints=extra_constraints):
             # all constraints are satisfied
             return ()
@@ -246,7 +317,7 @@ class FullFrontend(ConstrainedFrontend):
     # Serialization and such.
     #
 
-    def downsize(self):
+    def downsize(self) -> None:
         ConstrainedFrontend.downsize(self)
         self._tls.solver = None
         self._to_add = []
@@ -255,7 +326,7 @@ class FullFrontend(ConstrainedFrontend):
     # Merging and splitting
     #
 
-    def merge(self, others, merge_conditions, common_ancestor=None):
+    def merge(self: T, others, merge_conditions, common_ancestor=None) -> Tuple[bool, T]:
         return (
             self._solver_backend.__class__.__name__ == "BackendZ3",
             ConstrainedFrontend.merge(self, others, merge_conditions, common_ancestor=common_ancestor)[1],
